""" 
Taken from:
# https://stackoverflow.com/questions/43440813/shortest-great-circle-distance-between-a-point-and-a-polygon-on-a-sphere-globe
"""

from shapely.geometry import Polygon, mapping
import shapely
import numpy as np
import math
from numpy import linalg as LA

def Pairwise(iterable):
    """Iterate through an itertable returning adjacent pairs.
    
    Args:
        iterable (array-like): An iterable
    
    Returns:
        tuple: Pairs of sequential, adjacent entries from the iterable
    """
    it = iter(iterable)
    a = next(it, None)
    for b in it:
        yield (a, b)
        a = b


def LatLonToXYZ(lat,lon,radius):
    """Convert a latitude-longitude pair to 3D-Cartesian coordinates.

    Args:
        lat (float): Latitude in degrees
        lon (float): Longitude in degrees
        radius (float): Radius in arbitrary units
    
    Returns:
        tuple: x,y,z coordinates in arbitrary units
    """
    lat = np.radians(lat)
    lon = np.radians(lon)
    x = radius * np.cos(lon) * np.cos(lat)
    y = radius * np.sin(lon) * np.cos(lat)
    z = radius * np.sin(lat)
    return x,y,z


def XYZtoLatLon(x,y,z):
    """Convert 3D-Cartesian coordinates to a latitude-longitude pair.

    Args:
        x (float): x-coordinate in arbitrary units
        y (float): y-coordinate in arbitrary units
        z (float): z-coordinate in arbitrary units
    
    Returns:
        tuple: lat,lon pair in degrees
    """
    radius = np.sqrt(x*x+y*y+z*z)
    lat = np.degrees(np.arcsin(z/radius))
    lon = np.degrees(np.arctan2(y, x))   
    return lat,lon


def Haversine(lat1, lon1, lat2, lon2):
    """
    Calculate the Great Circle distance on Earth between two latitude-longitude
    points
    Args:
        lat1 (float): Latitude of Point 1 in degrees
        lon1 (float): Longtiude of Point 1 in degrees
        lat2 (float): Latitude of Point 2 in degrees
        lon2 (float): Longtiude of Point 2 in degrees
    Returns:
        float: Distance between the two points in kilometres
    """
    Rearth = 6371
    lat1 = np.radians(lat1)
    lon1 = np.radians(lon1)
    lat2 = np.radians(lat2)
    lon2 = np.radians(lon2)
    #Haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = np.sin(dlat/2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2)**2
    c = 2 * np.arcsin(np.sqrt(a)) 
    return Rearth*c


#http://stackoverflow.com/a/1302268/752843
def NearestPointOnGC(alat1,alon1,alat2,alon2,plat,plon):
    """
    Calculate the location of the nearest point on a Great Circle to a query point
    Args:
        lat1 (float): Latitude of start of arc in degrees
        lon1 (float): Longtiude of start of arc in degrees
        lat2 (float): Latitude of end of arc in degrees
        lon2 (float): Longtiude of end of arc in degrees
        plat (float): Latitude of query point in degrees
        plon (float): Longitude of query point in degrees
    Returns:
        tuple: A (lat,lon) pair in degrees of the closest point
    """
    Rearth    = 6371 #km
    #Convert everything to Cartesian coordinates
    a1 = np.array(LatLonToXYZ(alat1,alon1,Rearth))
    a2 = np.array(LatLonToXYZ(alat2,alon2,Rearth))
    p = np.array(LatLonToXYZ(plat, plon, Rearth))
    G = np.cross(a1,a2) #Plane of the Great Circle containing A and B
    F = np.cross(p,G)   #Plane perpendicular to G that passes through query pt
    T = np.cross(G,F)   #Vector marking the intersection of these planes
    T = Rearth*T/LA.norm(T) #Normalize to lie on the Great Circle
    tlat,tlon = XYZtoLatLon(*T)
    return tlat,tlon


def DistanceToGCArc(alat,alon,blat,blon,plat,plon):
    """
    Calculate the distance from a query point to the nearest point on a
    Great Circle Arc
    Args:
        lat1 (float): Latitude of start of arc in degrees
        lon1 (float): Longtiude of start of arc in degrees
        lat2 (float): Latitude of end of arc in degrees
        lon2 (float): Longtiude of end of arc in degrees
        plat (float): Latitude of query point in degrees
        plon (float): Longitude of query point in degrees
    Returns:
        float: The distance in kilometres from the query point to the great circle arc
    """
    tlat,tlon = NearestPointOnGC(alat,alon,blat,blon,plat,plon) #Nearest pt on GC
    abdist = Haversine(alat,alon,blat,blon)  #Length of arc
    atdist = Haversine(alat,alon,tlat,tlon)  #Distance arc start to nearest pt
    tbdist = Haversine(tlat,tlon,blat,blon)  #Distance arc end to nearest pt
    #If the nearest point T on the Great Circle lies within the arc, then the
    #length of the arc is approximately equal to the distance from T to each end
    #of the arc, accounting for floating-point errors
    PRECISION = 1e-3 #km 
    #We set the precision to a relatively high value because super-accuracy is not
    #to needed here and a failure to catch this can lead to vast under-estimates
    #of distance
    if np.abs(abdist-atdist-tbdist)<PRECISION: #Nearest point was on the arc
        return Haversine(tlat,tlon,plat,plon)
    #Okay, the nearest point wasn't on the arc, so the nearest point is one of the
    #ends points of the arc
    apdist = Haversine(alat,alon,plat,plon)
    bpdist = Haversine(blat,blon,plat,plon)
    return min(apdist,bpdist)


def Distance3dPointTo3dPolygon(lat,lon,geom):
    """
    Calculate the closest distance between a latitude-longitude query point
    and a `shapely` polygon defined by latitude-longitude points using only 
    spherical mathematics.

    Args:
        lat (float): Latitude of query point in degrees
        lon (float): Longitude of query point in degrees
        geom (`shapely` geometry): A `shapely` geometry whose points are in 
            latitude-longitude space
    Returns:
        dist (float): The minimum distance in kilometres between the polygon 
            and the query point.
    """
    if geom.type.values[0] == 'Polygon':
        dist = math.inf
#        xy = features[0]['geometry'][0].exterior.xy
        xy = mapping(geom)["features"][0]["geometry"]["coordinates"][0]
        #Polygons are closed rings, so the first-last pair is automagically delt with
        for p1, p2 in Pairwise(zip(*xy)):
            dist = min(dist,DistanceToGCArc(p1[1],p1[0],p2[1],p2[0],lat,lon))
    elif geom.type == 'MultiPolygon':
        dist = min(*[Distance3dPointTo3dPolygon(lat,lon,part) for part in geom])
    return dist

###############################################################################

# Jo wrote this

def Distance3dPointTo3dCoords(lat, lon, polylat, polylon):
    """
    Calculate the closest distance between a latitude-longitude query point
    and a `shapely` polygon defined by latitude-longitude points using only 
    spherical mathematics, modified from Distance3dPointTo3dPolygon.
    
    Args:
        lat (float): Latitude of query point in degrees
        lon (float): Longitude of query point in degrees
        polylat (array): Latitude borders polygon 
        polylon (array): Longitude borders polygon 
        geom (`shapely` geometry): A `shapely` geometry whose points are in 
            latitude-longitude space
    Returns:
        dist (float): The minimum distance in kilometres between the polygon 
            and the query point.
    """
    
    dist = math.inf
    for i in range(len(polylat)):
        if i > len(polylat) - 2:
            break
        dist = min(dist, DistanceToGCArc(polylat[i], polylon[i], 
                    polylat[i+1], polylon[i+1], lat, lon))
    
    return dist


def saa_distance_shapely(lat, lon, geom):
    """
    Wrapper to calculate distance to shapely geometry (defined in
    latitude-longitude space) for a series of points.
    
    Args:
        lat (float): Latitude of query point in degrees
        lon (float): Longitude of query point in degrees
        geom (`shapely` geometry): A `shapely` geometry whose points are in 
            latitude-longitude space
    Returns:
        dist (float): The minimum distance in kilometres between the polygon 
            and the query point.
    """
    dists = []
    for i in range(len(lat)):
        dist = Distance3dPointTo3dPolygon(lat[i], lon[i], geom)
        dists.append(dist)
    return dists


def saa_distance(lat, lon, poly_lat, poly_lon):
    """
    Wrapper to calculate distance to polygon (defined in
    latitude-longitude space) for a series of points.
    
    Args:
        lat (float): Latitude of query point in degrees
        lon (float): Longitude of query point in degrees
        polylat (array): Latitude borders polygon 
        polylon (array): Longitude borders polygon 
    Returns:
        dist (float): The minimum distance in kilometres between the polygon 
            and the query point.
    """
    dists = []
    for i in range(len(lat)):
        dist = Distance3dPointTo3dCoords(lat[i], lon[i], poly_lat, poly_lon)
        dists.append(dist)
    return dists



import os
import yaml
import numpy as np
import matplotlib.pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "niceplot.mplstyle")
plt.style.use(stylesheet)
from scipy.stats import norm
from sqlalchemy.orm import load_only
from timeit import default_timer as timer
import pandas as pd
from collections import OrderedDict,defaultdict

from acdc.database.connect_db import load_connection
from acdc.database import darkevents_schema
from acdc.utils.utils import sql_to_df

HV = {"FUVA": [163, 167, 169, 171, 173, 178],
      "FUVB": [163, 167, 169, 175]}
ALL_PHAS = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30]
DBNAME = "dark_events"
TIMING = True

def query_data(segment="*", mjdstart=0, mjdend=99999, x0=0, x1=16385, 
               y0=0, y1=1025, hv=163, getall=True):

    # Connect to database
    session, engine = load_connection("dark_events")
    
    # Define columns to return from database
    cols = ["pha", "hv", "segment"]

    # Get Tables for all HVs
#    possible_hvs = [163, 167, 169, 171, 173, 175, 178]
    possible_hvs = [171, 173]
    data = {}
    for hv in possible_hvs:
        tablename = f"DarkEventsHv{hv}"
        events_table = getattr(darkevents_schema, tablename)
        # Execute query
        query0 = timer()
        if getall is True:
            query = session.query(events_table).options(load_only(*cols))
        else:
            query = session.query(events_table).options(load_only(*cols))\
                    .filter(events_table.segment == segment)\
                    .filter(events_table.hv == hv)\
                    .filter(events_table.mjd > mjdstart)\
                    .filter(events_table.mjd < mjdend)\
                    .filter(events_table.xcorr > x0)\
                    .filter(events_table.xcorr < x1)\
                    .filter(events_table.ycorr > y0)\
                    .filter(events_table.ycorr < y1)
        results = query.all()
        query1 = timer()

        # Convert query results (list of attributes) to pandas dataframe
        df = sql_to_df(results, cols)
        data[hv] = df
        processing1 = timer()
        if TIMING is True:
            print(f"Query for HV={hv} took {query1-query0:.3g} seconds")
            print(f"Data formatting for HV={hv} took {processing1-query1:.3g} seconds")

    return data

def bin_data(data, segment, interactive=False):
    x = []
    y = []
    hvs_gain = OrderedDict()
    for hv in HV[segment]:
        df = data[hv] 
        phas = df.pha.values
        s_phas = list(set(phas))
        present = [x for x in ALL_PHAS if x in s_phas]
        if set(present) != set(ALL_PHAS):
            print(f"HV={hv} does not have counts at all PHAs")
            continue
        if interactive is True:
            plt.hist(phas, bins=np.arange(0,32))
            plt.show()
            x0 = int(input("Enter starting X for normal fit: "))
            x1 = int(input("Enter ending X for normal fit: "))
            plt.clf()
        else:
            x0 = 5
            x1 = 15
        subset = np.where((phas > x0) & (phas < x1))
        mean,std = norm.fit(phas[subset])
        modal_gain = mean
        hvs_gain[hv] = modal_gain
    hvs_keys = list(hvs_gain.keys())
    for i,hv in enumerate(hvs_keys):
        if i == 0:
            continue
        previous_hvs = np.arange(i)
        for j in previous_hvs:
            hv_diff = hv - hvs_keys[j] 
            gain_diff = hvs_gain[hv] - hvs_gain[hvs_keys[j]]
            relation = gain_diff / hv_diff
            starting_gain = hvs_gain[hvs_keys[j]]
            x.append(starting_gain)
            y.append(relation)
    
    return x,y

def plot_data(x, y, segment):
    fig, ax = plt.subplots(figsize=(24,12))
    ax.plot(x, y, "k,")
    ax.set_title("f{segment}")
    ax.set_xlabel("Starting Modal Gain")
    ax.set_ylabel("Pulse Height Bins per HV Step")
    out = f"{segment}_pha_hv_relation.png"
    plt.savefig(out, bbox_inches="tight")
    print(f"Wrote {out}")



from math import * 
import numpy as np
from scipy.optimize import minimize
from scipy.special import factorial
import matplotlib as mpl
import matplotlib.pyplot as plt
plt.rc('font', family='serif')
mpl.rcParams.update({'font.size': 12})
mpl.rcParams.update({'legend.labelspacing':0.25, 'legend.fontsize': 12})
mpl.rcParams.update({'errorbar.capsize': 4})


## Exponential distribution
def exp_fun (E, k, Et):

	val = k * np.exp (-E / Et) #/ (Et *(1.0 - exp(-31.0/Et)) )

	return val

## Exponential + normal distribution
def exp_norm_fun (E, k, Et, w, mu, sigma):

	val = k *np.exp (-E / Et) + w * np.exp (-np.power(E - mu, 2.0) / 2./np.power(sigma,2.0)) #/ sqrt(2.*pi) / sigma

	return val

## Likelihood in form of C-statistics for exponential distribution
def Cstat (param, data):

	k, Et = param

	E = np.linspace (0,31, 32)


	model = exp_fun (E, k, Et)

	C = 2 * (model - data * np.log(model) + np.log(factorial(data)))

	C = np.sum(C)

	return C

## Likelihood in form of C-statistics for exponential + normal distribution
def Cstat_alt (param, data):

	k, Et, w, mu, sigma = param


	E = np.linspace (0,31, 32)

	model = exp_norm_fun (E, k, Et, w, mu, sigma)

	C = 2 * (model - data * np.log(model) + np.log(factorial(data)))

	C = np.sum(C)

	if isnan(C):

		C = 5e12
	


	return C


## Function which determines which of two models describe the spectra the best:
## (1) - exponential distribution, in this case the function returns two parameters: k, Et
## (2) - exponential + normal distribution, in this case the function returns five parameters: k, Et for exponential
## and w, mu and sigma for normal distribution

def study_spectra (data):

	x0 = (2, 10)

	res = minimize(Cstat, x0, method='L-BFGS-B', tol=1e-2, args=(data), bounds=((0, np.max(data)), (0.01, 60))   )

	Cstat_A = res.fun
	A_par   = res.x
	
	
	x0 = (2, 10, 20, 10, 3)

	res = minimize(Cstat_alt, x0, method='L-BFGS-B', tol=1e-2, args=(data), bounds=((0, np.max(data)), (0.01, 60), (0, np.max(data)), (3,31), (0.01,60)   ))

	Cstat_B = res.fun
	B_par   = res.x


	if Cstat_B < Cstat_A - 2*3: ##  AIC - model B has three more parameters

		return B_par

	else:	
	
		return A_par


if __name__ == "__main__":

	print ('Test: ',  factorial(0, exact=False))

	E = np.linspace (0,31, 32)

	k = 30.0

	Et = 7.0

	#data_pseudo = exp_fun (E, k, Et)

	data_pseudo = exp_norm_fun (E, k, Et, 15, 12, 5)

	for i in range (0, len(E)):

		data_pseudo[i] = int(data_pseudo[i])

		print (i, data_pseudo[i])


	
	

	val = study_spectra (data_pseudo)

	if len(val) < 3:

		model = exp_fun (E, val[0], val[1])

	else:

		model = exp_norm_fun (E, val[0], val[1], val[2], val[3], val[4])


	plt.plot (E, data_pseudo, 'k-', drawstyle='steps-mid', label='Pseudo data')
	plt.plot (E, model,       'b--', drawstyle='steps-mid',label='Model')
	plt.xlabel ('E')
	plt.ylabel ('Counts')
	plt.legend()
	plt.show()







import sqlite3
from sqlite3 import Error
import sys
from math import * 
import numpy as np
from scipy.optimize import minimize
from scipy.special import factorial
import matplotlib as mpl
import matplotlib.pyplot as plt
plt.rc('font', family='serif')
mpl.rcParams.update({'font.size': 12})
mpl.rcParams.update({'legend.labelspacing':0.25, 'legend.fontsize': 12})
mpl.rcParams.update({'errorbar.capsize': 4})

## Connecting to the database
def create_connection(db_file):
	""" create a database connection to a SQLite database """
	conn = None
	try:
		conn = sqlite3.connect(db_file)
		print(sqlite3.version)
	except Error as e:
		print(e)
	return conn

def select_all_tasks(conn, hv, segment):
	cur = conn.cursor()
	cur.execute("SELECT id, latitude, longitude, expstart, region, solar_flux, dark_pha0, dark_pha1, dark_pha2, dark_pha3, \
                 dark_pha4,  dark_pha5,  dark_pha6,  dark_pha7,  dark_pha8,  dark_pha9,  dark_pha10, dark_pha11, \
                 dark_pha12, dark_pha13, dark_pha14, dark_pha15, dark_pha16, dark_pha17, dark_pha18, dark_pha19, \
                 dark_pha20, dark_pha21, dark_pha22, dark_pha23, dark_pha24, dark_pha25, dark_pha26, dark_pha27, \
                 dark_pha28, dark_pha29, dark_pha30, dark_pha31 \
                 FROM Darks  \
                 WHERE region='inner' and segment='"+str(segment)+"'  and hv="+str(hv)+' and id > 0 and id < 50')


	rows = cur.fetchall()

	data = np.zeros(32)

	for row in rows:

		print (row)

		for i in range (0, 32):
	
			data[i] = data[i] + float(row[i+6])

		#break

	return data



## Exponential distribution
def exp_fun (E, k, Et):

	val = k * np.exp (-E / Et) #/ (Et *(1.0 - exp(-31.0/Et)) )

	return val

## Exponential + normal distribution
def exp_norm_fun (E, k, Et, w, mu, sigma):

	val = k *np.exp (-E / Et) + w * np.exp (-np.power(E - mu, 2.0) / 2./np.power(sigma,2.0)) #/ sqrt(2.*pi) / sigma

	return val

## Likelihood in form of C-statistics for exponential distribution
def Cstat (param, data):

	k, Et = param

	E = np.linspace (0,31, 32)


	model = exp_fun (E, k, Et)

	C = 2 * (model - data * np.log(model) + np.log(factorial(data)))

#	for i in range (0, len(C)):
#		print (i, C[i])

#	print ('.. ', np.sum(C), np.sum(C[2:29]))

#	sys.exit(0)	

	C = np.sum(C[2:29])

	

	#print ('---', k, Et, C)

	return C

## Likelihood in form of C-statistics for exponential + normal distribution
def Cstat_alt (param, data):

	k, Et, w, mu, sigma = param


	E = np.linspace (0,31, 32)

	model = exp_norm_fun (E, k, Et, w, mu, sigma)

	C = 2 * (model - data * np.log(model) + np.log(factorial(data)))

	C = np.sum(C[2:29])

	print ('---', k, Et, w, mu, sigma, C)

	if isnan(C):

		C = 5e12
	


	return C


## Function which determines which of two models describe the spectra the best:
## (1) - exponential distribution, in this case the function returns two parameters: k, Et
## (2) - exponential + normal distribution, in this case the function returns five parameters: k, Et for exponential
## and w, mu and sigma for normal distribution

def study_spectra (data):

	x0 = (np.max(data), 10)

	res = minimize(Cstat, x0, method='L-BFGS-B', tol=1e-6, args=(data), bounds=((0, 2.0*np.max(data)), (0.01, 60))   )

	Cstat_A = res.fun
	A_par   = res.x
	
	
	x0 = (A_par[0], A_par[1], A_par[0]/3.0, 10, 3)

	res = minimize(Cstat_alt, x0, method='L-BFGS-B', tol=1e-6, args=(data), bounds=((0, 2.0*np.max(data)), (0.01, 60), (0, np.max(data)), (3,31), (0.01,60)   ))

	Cstat_B = res.fun
	B_par   = res.x

	print ('Cstat for A: ', Cstat_A, ' Cstat for B: ', Cstat_B)

	if Cstat_B < Cstat_A - 2*3: ##  AIC - model B has three more parameters

		return B_par

	else:	
	
		return A_par


if __name__ == "__main__":

	hv = 167
	segment = 'FUVA'


	conn = create_connection('cos_dark.db')
	data = select_all_tasks(conn, hv, segment)
	
	print ('-------------------------------------')
	
	for i in range (0, len(data)):

		print  (i, data[i])

	print ('-------------------------------------')

	#sys.exit(0)

	#print ('Test: ',  factorial(0, exact=False))

	E = np.linspace (0,31, 32)

	#k = 30.0

	#Et = 7.0

	#data_pseudo = exp_fun (E, k, Et)

	#data_pseudo = exp_norm_fun (E, k, Et, 15, 12, 5)

	#for i in range (0, len(E)):

	#	data_pseudo[i] = int(data_pseudo[i])

	#	print (i, data_pseudo[i])


	
	

	val = study_spectra (data)

	print ('Result of optimisation: ', val)

	if len(val) < 3:

		model = exp_fun (E, val[0], val[1])

	else:

		model = exp_norm_fun (E, val[0], val[1], val[2], val[3], val[4])


	plt.plot (E, data, 'k-', drawstyle='steps-mid', label='Data')
	plt.plot (E, model,       'b--', drawstyle='steps-mid',label='Model')
	plt.xlabel ('E')
	plt.ylabel ('Counts')
	plt.legend()
	plt.show()







"""
Perform a custom dark subtraction on COS corrtag files.
"""

import os
import argparse
import copy
import asdf
from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "analysis", "niceplot.mplstyle")
plt.style.use(stylesheet)
import glob

from .predict_dark_level import bin_science, get_binning_pars

RESEL = [6, 10]
PHA_INCLUSIVE = [2, 23]
PHA_INCL_EXCL = [2, 24]

def subtract_dark(corrtags, datadir, fact=1, outdir=".", overwrite=False):
    """Perform the custom dark correction.

    Args:
        corrtags (list or array-like): Input, unmodified science corrtags.
        datadir (str): Directory that houses the modeled dark rates for each
            input corrtag.
        fact (int): Changes resolution, 1=highest resolution, 8=lowest resolution. 
            To keep at same input resolution, use 1.
        outdir (str): Directory where custom corrtags will be written.
        """

    for item in corrtags:
        try:
            acdc_done = fits.getval(item, "ACDCCORR")
            if acdc_done == "COMPLETE":
                print(f"WARNING:\nCustom dark correction has already been applied to {item}, skipping...")
                continue
        except KeyError:
            pass
        
        outfile = os.path.join(outdir, f"corrected_{os.path.basename(item)}")
        if os.path.exists(outfile) and overwrite is False:
             print(f"WARNING: Corrected corrtag {outfile} already exists and overwrite is False, skipping...")
             continue
        if not os.path.exists(outdir):
            os.makedirs(outdir)
       
        hdr0 = fits.getheader(item, 0) 
        rootname = hdr0["rootname"]
        segment = hdr0["segment"]
        cenwave = hdr0["cenwave"]
        lp = hdr0["life_adj"]
        pred_noise_file = os.path.join(datadir, f"{rootname}_{segment}_noise_complete.asdf")
        pred_noise_af = asdf.open(pred_noise_file)
        pha_str = f"pha{PHA_INCL_EXCL[0]}-{PHA_INCL_EXCL[1]}"
        pred_noise = pred_noise_af[pha_str]
        b = get_binning_pars(pred_noise_af)
        binned, nevents = bin_science(item, b, segment, cenwave, lp, fact, exclude_airglow=False)
        
        hdulist = fits.open(item)
        data = hdulist[1].data
        
        xstart = b["xstart"]
        xend = b["xend"]
        ystart = b["ystart"]
        yend = b["yend"]
        phastart = b["phastart"]
        phaend = b["phaend"]
        bin_x = b["bin_x"]
        bin_y = b["bin_y"]

        #wwf = open("tst.txt", "w")
        logic = np.zeros((binned.shape[0], binned.shape[1]))
       
        all_delta_eps_ind = []
        for i in range(len(data["xcorr"])):
            if data["xcorr"][i] > xstart and data["xcorr"][i] < xend:
                if data["ycorr"][i] > ystart and data["ycorr"][i] < yend:
                    if data["pha"][i] > phastart and data["pha"][i] < phaend:
                        x = int((data["xcorr"][i] - xstart) // bin_x * fact)
                        y = int((data["ycorr"][i] - ystart) // bin_y)
                        delta_eps = pred_noise[y, int(x/fact)] / float(fact)
                        logic[y,x] = 1
                        delta_eps_ind = delta_eps / float(nevents[y,x])
                        if np.isnan(delta_eps_ind):
                            delta_eps_ind = 0
        #                outstr = f"{data['epsilon'][i]}\t{delta_eps}\t{delta_eps_ind}\n"
                        data["epsilon"][i] -= delta_eps_ind
        #                wwf.write(outstr)
        #wwf.close()

        #plt.imshow(logic, aspect="auto", origin="lower")
        #plt.savefig("logic.png")
        #plt.clf()
        
#        hdulist = fits.open(item, mode="update")
#        data = hdulist[1].data
#        phainds = np.where((data["pha"] >= b["phastart"]) &
#                           (data["pha"] <= b["phaend"]))
#        phadata = data[phainds]
#        innerinds = np.where((phadata["xcorr"] > b["xstart"]) &
#                             (phadata["xcorr"] < b["xend"]) &
#                             (phadata["ycorr"] > b["ystart"]) &
#                             (phadata["ycorr"] < b["yend"]))
#        filtered = phadata[innerinds]
#
#        wwf = open("tst.txt", "w")
#        logic = np.zeros((binned.shape[0], binned.shape[1]))
##        xs = (filtered["xcorr"] - b["xstart"]) // b["bin_x"] * fact
##        xs = (filtered["ycorr"] - b["ystart"]) // b["bin_y"]
##        delta_eps = pred_noise[
#        print("before loop 1")
#        all_delta_eps_ind = []
#        for i in range(len(filtered["xcorr"])):
#            print(i)
#            x = int((filtered["xcorr"][i] - b["xstart"]) // b["bin_x"] * fact)
#            y = int((filtered["ycorr"][i] - b["ystart"]) // b["bin_y"])
#            delta_eps = pred_noise[y, int(x/fact)] / float(fact)
#            logic[y,x] = 1
#            delta_eps_ind = delta_eps / float(nevents[y,x])
##            all_delta_eps_ind.append(delta_eps_ind)
#            outstr = f"{data[phainds][innerinds][i]}\t{delta_eps}\t{delta_eps_ind}\n"
#        #    data[phainds][innerinds]["epsilon"][i] -= delta_eps_ind
#            filtered["epsilon"][i] -= delta_eps_ind
#            wwf.write(outstr)
#        
#        data[phainds][innerinds]["epsilon"] = all_delta_eps_ind
##        data[phainds][innerinds]["epsilon"] -= all_delta_eps_ind
#        print("after loop 1")
#        wwf.close()
#
#        plt.imshow(logic, aspect="auto", origin="lower")
#        plt.savefig("logic.png")
#        plt.clf()

        
        inds = np.where(logic < 1)
        if len(logic[inds]) != 0:
            if len(logic[inds]) >  len(data):
                #num = int(np.ceil(len(logic[inds]) / len(data)))
                #tmp = [data]*num
                #longerdata = np.concatenate(tmp)
                longerhdu = fits.BinTableHDU.from_columns(hdulist[1].columns, nrows=len(logic[inds]))
                for colname in hdulist[1].columns.names:
                    longercol = np.resize(data[colname], len(logic[inds]))
                    longerhdu.data[colname][:] = longercol
                new_events = longerhdu.data
            else:
                new_events = copy.deepcopy(data[len(data) - len(inds[0]):])
            for i in range(len(inds[0])):
                y = b["bin_y"] * inds[0][i] + b["ystart"] 
                x = b["bin_x"] * inds[1][i] + b["xstart"]
                new_events["rawx"][i] = x 
                new_events["rawy"][i] = y
                new_events["xcorr"][i] = x 
                new_events["ycorr"][i] = y
                new_events["xfull"][i] = x 
                new_events["yfull"][i] = y
                new_events["pha"][i] = 10
                new_events["dq"][i] = 0
                new_events["epsilon"][i] = -pred_noise[inds[0][i], int(inds[1][i]/fact)] / fact
            data = np.concatenate([data, new_events])
        nans = np.isnan(data["EPSILON"])
        data["EPSILON"][nans] = 0
        hdulist[1].data = data
        hdulist.writeto(outfile, overwrite=overwrite)
        with fits.open(outfile, mode="update") as hdulist:
            hdr0 = hdulist[0].header
            hdr0.set("ACDCCORR", "COMPLETE")
        print(f"Wrote {outfile}")

        pred_noise_af.close() 

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--datadir",
                        help="Path to science corrtags")
    parser.add_argument("-o", "--outdir", default=".",
                        help="Output directory")
    args = parser.parse_args()
    corrtags = glob.glob(os.path.join(args.datadir, "*corrtag*fits"))

    subtract_dark(corrtags, args.datadir, outdir=args.outdir)




import os
os.environ["GOTO_NUM_THREADS"] = "5"
os.environ["OMP_NUM_THREADS"] = "5"
os.environ["OPENBLAS_NUM_THREADS"] = "5"
os.environ["CRDS_SERVER_URL"] = "https://hst-crds.stsci.edu"
import crds
import datetime
import argparse
import glob
import copy
import asdf
from astropy.io import fits
from astropy.table import Table
from calcos import ccos
import numpy.ma as ma
import numpy as np
import pandas as pd
import dask
import matplotlib.pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "../analysis", "niceplot.mplstyle")
plt.style.use(stylesheet)
from matplotlib import ticker
from matplotlib.backends.backend_pdf import PdfPages
from astropy.convolution import Gaussian2DKernel, interpolate_replace_nans

from acdc.database.query_cos_dark import files_by_mjd
from acdc.utils.utils import get_psa_wca, unbin_coords, bin_coords

class Superdark():
    def __init__(self, hv, segment, mjdstarts, mjdends, dayint=100, bin_x=1,
                 bin_y=1, bin_pha=1, phastart=1, phaend=31, pha_bins=None, 
                 gsagtab=None, bpixtab=None, region="inner", outfile=None,
                 outdir=".", xylimits=None, overwrite=False, sdqflags=8346):
                 #gsagtab="/grp/hst/cdbs/lref/71j1935gl_gsag.fits", 
                 #bpixtab="/grp/hst/cdbs/lref/69m20528l_bpix.fits", 
        """
        To keep things consistent, start and stop range are defined the same
        as within python- start is inclusive but stop is EXCLUSIVE. E.g.
        using a PHA range of 1 to 31 means you will get PHA=1 up to PHA=30, but
        not PHA=31.
        """
        self.segment = segment
        self.hv = hv
        self.region = region
        self.mjdstarts = mjdstarts
        self.mjdends = mjdends
        self.ndays = [mjdends[i] - mjdstarts[i] for i in range(len(mjdstarts))]
        self.dayint = dayint
        self.bin_x = bin_x
        self.bin_y = bin_y
        self.overwrite = overwrite
        self.sdqflags = sdqflags

        self.phastart = phastart
        self.phaend = phaend
        if bin_pha == "all":
            bin_pha = phaend - phastart
        self.bin_pha = bin_pha
        self.pha_bins = pha_bins
        self.get_pha_bins()

        self.superdarks = []
        if gsagtab is None:
            gsagtab = self.get_reffile("gsagtab")
        self.gsagtab = gsagtab
        if bpixtab is None:
            bpixtab = self.get_reffile("bpixtab")
        self.bpixtab = bpixtab
        self.outfile = outfile
        self.outdir = outdir

        if xylimits == None:
            self.get_xy_limits()

        self.gsag_holes = get_gsag_holes(self.gsagtab, self.segment, self.hv)
        self.fixed_gsag = False 
        self.is_binned = False
#        self.get_bpix_regions()


    @classmethod
    def from_asdf(cls, superdark, overwrite=False):
        af = asdf.open(superdark)
        if "pha_bins" in af.keys():
            pha_bins = copy.deepcopy(af["pha_bins"])
        else:
            pha_bins = None
        self = cls(hv=af["hv"], segment=af["segment"], mjdstarts=af["mjdstarts"], 
                   mjdends=af["mjdends"], bin_x=af["bin_x"], bin_y=af["bin_y"],
                   bin_pha=af["bin_pha"], phastart=af["phastart"],
                   phaend=af["phaend"], pha_bins=pha_bins, xylimits="predetermined", 
                   outfile=superdark)
        for k in ["bin_pha", "bin_x", "bin_y", "xstart", "xend", "ystart", 
                  "yend", "phastart", "phaend"]:
            setattr(self, k, af[k])

        self.outfile = superdark
        self.outfile_basename = os.path.basename(superdark)
        self.total_exptime = af["total_exptime"]
        self.total_files = af["total_files"]
        try:
            self.fixed_gsag = af["fixed_gsag"]
        except KeyError:
            self.fixed_gsag = False
        try:
            self.is_binned = af["is_binned"]
        except KeyError:
            if "binned" in os.path.basename(superdark):
                self.is_binned = True
            else:
                self.is_binned = False
        self.overwrite = overwrite

        self.pha_images = {}
        try:
            self.dq_image = copy.deepcopy(af["dq_image"])
        except KeyError:
            sh = af[f"pha{pha_bins[0]}-{pha_bins[1]}"].shape
            self.dq_image = np.zeros(sh[0]*sh[1]).reshape(sh).astype(int)
        for i in range(len(self.pha_bins)-1):
            key = f"pha{self.pha_bins[i]}-{self.pha_bins[i+1]}"
            self.pha_images[key] = copy.deepcopy(af[key])
            self.superdarks.append(copy.deepcopy(af[key]))

        try:
            self.creation_date = af["creation_date"]
        except KeyError:
            self.creation_date = "unknown"
        af.close()
        return self


    def get_reffile(self, filetype):
        now = datetime.datetime.now()
        current_pmap = crds.get_default_context()
        config = {"INSTRUME": "COS",
                  "CENWAVE": 1291,
                  "DETECTOR": "FUV",
                  "DATE-OBS": now.strftime("%Y-%m-%d"),
                  "TIME-OBS": "00:00:00"}
        crds_results = crds.getrecommendations(parameters=config, reftypes=[filetype],
                                               context=current_pmap, observatory="hst")
        reffile0 = crds_results[filetype]
        try:
            lref = os.environ["lref"]
        except KeyError as e:
            print(e.message)
            print("You must define the $lref environment variable- this is where all COS reference files are located")
        reffile = os.path.join(lref, reffile0)
        
        return reffile
    

    def get_pha_bins(self):
        if self.pha_bins is None:
            self.pha_bins = np.arange(self.phastart, self.phaend+1, self.bin_pha)
        else:
            self.phastart = self.pha_bins[0]
            self.phaend = self.pha_bins[-1]
            grid = list(set(self.pha_bins[1:] - self.pha_bins[:-1]))
            if len(grid) == 1:
                self.bin_pha = grid[0]
            else:
                self.bin_pha = None


    def get_xy_limits(self):
        if self.segment == "FUVA":
            if self.region == "inner":
                xstart, xend, ystart, yend = 1260, 15119, 375, 660
        elif self.segment == "FUVB":
            if self.region == "inner":
                xstart, xend, ystart, yend = 1000, 14990, 405, 740
        else:
            raise Exception(f"Invalid segment specified: {self.segment}")
        while (xstart // self.bin_x) * self.bin_x < xstart:
            xstart += 1
        while (ystart // self.bin_y) * self.bin_y < ystart:
            ystart += 1
        self.xstart, self.ystart = xstart, ystart
        self.xend = (xend // self.bin_x) * self.bin_x
        self.yend = (yend // self.bin_y) * self.bin_y
        self.bin_xstart = self.xstart // self.bin_x
        self.bin_ystart = self.ystart // self.bin_y
        self.bin_xend = self.xend // self.bin_x
        self.bin_yend = self.yend // self.bin_y


    def create_superdark(self, dark_df=None):

        @dask.delayed
        def _bin_dark(self, dark_df, phastart, phaend):
            pha_images = {} 
            darks = dark_df["fileloc"].values
            assert len(darks) != 0, "ERROR: no darks found at all!!"
                
            total_exptime = sum([fits.getval(x, "exptime", 1) for x in darks])
            # Handle gainsag holes
            gsag_df = self.gsag_holes.loc[(self.gsag_holes["DATE"] > self.mjdstarts[0]) & 
                                          (self.gsag_holes["DATE"] < self.mjdends[0])] 
            print(f"   Binning corrtags, {phastart} <= PHA < {phaend}...")
            sum_image = self.bin_corrtags(darks, phastart=phastart, phaend=phaend)

            ylen,xlen = sum_image.shape
            dq_image = np.zeros((ylen, xlen)).astype(int)
            
            # Loop through each gainsag hole and mark the DQ image as DQ=1
            for j in range(len(gsag_df)):
                dq_image[gsag_df.iloc[j]["Y0"]:gsag_df.iloc[j]["Y1"], 
                    gsag_df.iloc[j]["X0"]:gsag_df.iloc[j]["X1"]] = 1

            ## Handle bad pixel regions
            #bpix_df = self.bpix_regions
            #bpix_df = bpix_df.loc[(bpix_df.SDQ == True) & (bpix_df.SEGMENT == self.segment)]
            ## Loop through each bad pixel region and set affected pixels to a false value of 99999
            #for j in range(len(bpix_df)):
            #    sum_image[bpix_df.iloc[j]["Y0"]:bpix_df.iloc[j]["Y1"], 
            #        bpix_df.iloc[j]["X0"]:bpix_df.iloc[j]["X1"]] = 99999

            ## Handle hotspots
            #spot_df = self.hotspots
            #spot_df = spot_df.loc[(spot_df.SEGMENT == self.segment) & (spot_df.START > self.mjdstarts[0]) & (spot_df.STOP < self.mjdends[0])]
            ## Loop through each hotspot and set affected pixels to a false value of 99999
            #for j in range(len(spot_df)):
            #    sum_image[spot_df.iloc[j]["Y0"]:spot_df.iloc[j]["Y1"], spot_df.iloc[j]["X0"]:spot_df.iloc[j]["X1"]] = 99999

            tmp = sum_image.reshape(1024 // self.bin_y, self.bin_y,
                                    16384 // self.bin_x, self.bin_x)
            binned = tmp.sum(axis=3).sum(axis=1)
            binned_inner = binned[self.bin_ystart:self.bin_yend, self.bin_xstart:self.bin_xend]
            tmp_dq = dq_image.reshape(1024 // self.bin_y, self.bin_y,
                                    16384 // self.bin_x, self.bin_x)
            binned_dq = tmp_dq.sum(axis=3).sum(axis=1)
            binned_inner_dq = binned_dq[self.bin_ystart:self.bin_yend, self.bin_xstart:self.bin_xend]
            key = f"pha{phastart}-{phaend}"
            pha_images[key] = binned_inner
            return pha_images, total_exptime, binned_inner_dq

        runstart = datetime.datetime.now()
        print("\nStart time: {}".format(runstart))
        total_exptime = 0
        total_files = 0
        self.pha_images = {}
        if dark_df is None:
            notfilled = True
            dark_dfs = []
            total_days = 0
            for k in range(len(self.mjdstarts)):
                start = self.mjdstarts[k]
                done = False
                while done is False:
                    total_days += self.dayint
                    if total_days > self.ndays[k]:
                        total_days = self.ndays[k]
                        done = True
                    end = start + self.dayint
                    if end > self.mjdends[k]:
                        end = self.mjdends[k]
                    print(f"Using darks from MJD {start:,}-{end:,}; {total_days}/{self.ndays[k]} days")
                    print("   Querying...")
                    dark_df = files_by_mjd(start, end, segment=self.segment, hv=self.hv)
                    dark_dfs.append(dark_df)
                    print("   Query done")
                    start = end
            dark_df = pd.concat(dark_dfs)
        
        delayed_bin = [_bin_dark(self, dark_df, self.pha_bins[i], self.pha_bins[i+1]) for i in range(len(self.pha_bins)-1)]
        out = dask.compute(*delayed_bin, scheduler='multiprocessing', num_workers=15)
        print("   Binning done")
        for grp in out:
            dct = grp[0]
            dct_dq = grp[2]
            key = list(dct.keys())[0]
            self.pha_images[key] = dct[key]
        self.dq_image = out[0][2]
        total_exptime += out[0][1]

        ## Undo 99999 put in for gainsag holes
        #for i in range(len(self.pha_bins)-1):
        #    key = f"pha{self.pha_bins[i]}-{self.pha_bins[i+1]}"
        #    inds = np.where(self.pha_images[key] >= 99999)
        #    print(f"Zeroing out {len(inds[0])} gain sagged pixels for {key}")
        #    self.pha_images[key][inds] = 0
        #    inds = np.where(self.pha_images[key] >= 99999)
        #    assert len(inds[0]) == 0, f"Not all gain sag holes were zeroed for {key}"
        #    self.superdarks.append(self.pha_images[key])

        self.total_exptime = total_exptime
        self.total_files = len(dark_df["fileloc"].values)

        # Write out ASDF file
        self.write_superdark()
        runend = datetime.datetime.now()
        print("End time: {}".format(runend))
        print("Total time: {}".format(runend-runstart))

    
    def write_superdark(self, user_outfile=None, overwrite=False):
        data_dict = self.pha_images
        data_dict["xstart"] = self.xstart
        data_dict["ystart"] = self.ystart
        data_dict["phastart"] = self.phastart
        data_dict["xend"] = self.xend
        data_dict["yend"] = self.yend
        data_dict["phaend"] = self.phaend
        data_dict["bin_x"] = self.bin_x
        data_dict["bin_y"] = self.bin_y
        data_dict["bin_pha"] = self.bin_pha
        data_dict["pha_bins"] = self.pha_bins
        data_dict["mjdstarts"] = self.mjdstarts
        data_dict["mjdends"] = self.mjdends
        data_dict["segment"] = self.segment
        data_dict["hv"] = self.hv
        data_dict["total_exptime"] = self.total_exptime
        data_dict["total_files"] = self.total_files
        data_dict["fixed_gsag"] = self.fixed_gsag
        data_dict["is_binned"] = self.is_binned
        data_dict["dq_image"] = self.dq_image
        now = datetime.datetime.now()
        self.creation_date = now.strftime("%Y-%m-%d %H:%M")
        data_dict["creation_date"] = self.creation_date
        af = asdf.AsdfFile(data_dict)
        today = datetime.datetime.now().strftime("%d%b%y-%H:%M:%S")
        if self.outfile is None and user_outfile is None:
            self.outfile = f"superdark_{self.segment}_{self.hv}_{today}.asdf"
        elif self.outfile is None and user_outfile is not None:
            self.outfile = user_outfile
        if os.path.exists(self.outfile):
            assert sum([overwrite, self.overwrite]) > 0, f"Output superdark {self.outfile} already exists and overwrite is False"

        af.write_to(self.outfile)
        print(f"Wrote {self.outfile}")


    def fix_gsag(self, method="interpolate", kernel=None, row_threshold=.3,
                 lp1_interpolate=None):
        superdarks = []

        if self.segment == "FUVA":
            _, lp1_lims = bin_coords([0, 0], [459, 528], self.bin_x,
                self.bin_y, self.xstart, self.ystart)
        else:
            _, lp1_lims = bin_coords([0, 0], [521, 587], self.bin_x,
                self.bin_y, self.xstart, self.ystart)

        print(f"Looking for rows where >= {int(row_threshold*100)}% of the row is gain-sagged...")
        for i in range(len(self.pha_bins)-1):
            key = f"pha{self.pha_bins[i]}-{self.pha_bins[i+1]}"
            ylen,xlen = self.pha_images[key].shape
            for j in range(ylen):
                #if lp1_lims[0] <= j <= lp1_lims[1] and lp1_interpolate is True:
                #    print(f"    lp1_interpolate is True, interpolating over row {j}")
                #    self.dq_image[j,:] = 1
                #    continue
                if lp1_lims[0] <= j <= lp1_lims[1] and lp1_interpolate is False:
                    continue
                sagged = np.where(self.dq_image[j,:] == 1)
                frac_sagged = len(sagged[0])/xlen
                if frac_sagged >= row_threshold:
                    print(f"    {frac_sagged*100:.1f}% of row {j} is gain-sagged, entire row will be corrected")
                    self.dq_image[j,:] = 1

        inds = np.where(self.dq_image == 1)
        # Zero out gain sagged regions
        if method == "zero":
            print(f"Zeroing out gain sagged pixels...")
            for i in range(len(self.pha_bins)-1):
                key = f"pha{self.pha_bins[i]}-{self.pha_bins[i+1]}"
                self.pha_images[key][inds] = 0
                #inds = np.where(self.pha_images[key] >= 99999)
                #assert len(inds[0]) == 0, f"Not all gain sag holes were zeroed for {key}"
                superdarks.append(self.pha_images[key])
        elif method == "boost":
            print("Boosting gain sagged pixels...")
            for i in range(len(self.pha_bins)-1):
                key = f"pha{self.pha_bins[i]}-{self.pha_bins[i+1]}"
                indsbelow = (inds[0]-1, inds[1])
                indsabove = (inds[0]+1, inds[1])
                avg_around = (self.pha_images[key][indsbelow]+self.pha_images[key][indsabove])/2
                self.pha_images[key][inds] = avg_around
                superdarks.append(self.pha_images[key])
        elif method == "interpolate":
            print("Interpolating over gain sagged pixels...")
            if kernel is None:
                kernel = Gaussian2DKernel(x_stddev=.3, y_stddev=.3, x_size=11, y_size=5)
            for i in range(len(self.pha_bins)-1):
                key = f"pha{self.pha_bins[i]}-{self.pha_bins[i+1]}"
                self.pha_images[key][inds] = np.nan
                interp_sd = interpolate_replace_nans(self.pha_images[key], kernel)
                self.pha_images[key] = interp_sd
                superdarks.append(self.pha_images[key])
        self.superdarks = superdarks
        self.fixed_gsag = True 


    def get_hotspots(self, extnum):
        spot = Table.read(self.spottab, format="fits", hdu=extnum)
        spot_df = spot.to_pandas()
        spot_df = spot_df.rename(columns={"LX": "X0", "LY": "Y0"})
        spot_df["X1"] = spot_df["X0"] + spot_df["DX"]
        spot_df["Y1"] = spot_df["Y0"] + spot_df["DY"]
        spot_df = spot_df.astype("int32")
        str_df = spot_df.select_dtypes([np.object])
        str_df = str_df.stack().str.decode('utf-8').unstack()
        for col in str_df:
            spot_df[col] = str_df[col]

        self.hotspots = spot_df
    

    def get_bpix_regions(self):
        bpix = Table.read(self.bpixtab, format="fits", hdu=1)
        bpix_df = bpix.to_pandas()
        bpix_df = bpix_df.rename(columns={"LX": "X0", "LY": "Y0"})
        bpix_df["X1"] = bpix_df["X0"] + bpix_df["DX"]
        bpix_df["Y1"] = bpix_df["Y0"] + bpix_df["DY"]
        # This is to fix the astropy bytes string columns 
        str_df = bpix_df.select_dtypes([np.object])
        str_df = str_df.stack().str.decode('utf-8').unstack()
        for col in str_df:
            bpix_df[col] = str_df[col]
        bpix_df["SDQ"] =  np.where(bpix_df["DQ"]&self.sdqflags == bpix_df["DQ"], True, False)

        self.bpix_regions = bpix_df


    def bin_corrtags(self, corrtag_list, phastart, phaend, xtype='XCORR', ytype='YCORR', sdqflags=0):
        """Bin one or more corrtag event lists into a 2D image.

        Modifed from /user/jotaylor/git/jotools/utils.py

        Parameters
        ----------
        corrtag_list : str
            list of datasets to combine into an image
        xtype : str, optional
            X-coordinate type
        ytype : str, optional
            Y-coordinate type

        Returns
        -------
        image : np.ndarray
            2D image of the corrtag
        """

        final_image = np.zeros((1024, 16384)).astype(np.float32)

        for filename in corrtag_list:
            image = np.zeros((1024, 16384)).astype(np.float32)
            events = fits.getdata(filename, 1)
            inds = np.where((events["pha"] >= phastart) & (events["pha"] < phaend))
            events = events[inds]
            if not len(inds[0]):
                print(f"No counts for PHA>={phastart} and PHA<{phaend}")
                return image

            # Call for this is x_values, y_values, image to bin to, offset in x
            # ccos.binevents(x, y, array, x_offset, dq, sdqflags, epsilon)
            ccos.binevents(events[xtype].astype(np.float32),
                           events[ytype].astype(np.float32),
                           image,
                           0,
                           events['dq'],
                           sdqflags)

            final_image += image

        return final_image


    def bin_superdark(self, bin_x, bin_y, pha_bins=None, outfile=None, verbose=True, writefile=True, makeplots=True):
        
        # Bin across PHA
        self.pha_images = {}
        if pha_bins is not None:
            for b in pha_bins:
                if b not in self.pha_bins:
                    raise IndexError(f"Previous PHA bins not compatible with new bins, {self.pha_bins} vs {pha_bins}")
            superdarks = []
            inds = np.nonzero(np.in1d(self.pha_bins, pha_bins))[0]
            for i in range(len(inds)-1):
                superdark = self.superdarks[i]
                for j in range(i+1, inds[i+1]):
                    superdark += self.superdarks[j]
                superdarks.append(superdark)
            self.pha_bins = np.array(pha_bins)
            self.get_pha_bins()
            self.superdarks = superdarks
        
        infile = self.outfile
        if outfile is None and writefile is True:
            nowdt = datetime.datetime.now()
            now = nowdt.strftime("%d%b%Y")
            outfile = self.outfile.replace(".asdf", f"_binned_{now}.asdf")
        self.outfile = outfile
        if writefile is True and self.overwrite is False and os.path.exists(outfile):
            print(f"WARNING: Output superdark {outfile} already exists and overwrite is False, skipping...")
            return

        sh = np.shape(self.superdarks[0])
        xdim = sh[1]
        ydim = sh[0]
        b_x0 = 0
        b_y0 = 0
        b_x1 = (xdim // bin_x) * bin_x 
        b_y1 = (ydim // bin_y) * bin_y
        self.xend = b_x1 * self.bin_x + self.xstart
        self.yend = b_y1 * self.bin_y + self.ystart
# TODO
        # Should this really be *=, not just =?        
        self.bin_x *= bin_x
        self.bin_y *= bin_y
        # Bin DQ image
        binned_dq = self.dq_image[b_y0:b_y1, b_x0:b_x1]
        binned_sh_dq = np.shape(binned_dq)
        binned_xdim_dq = binned_sh_dq[1]
        binned_ydim_dq = binned_sh_dq[0]
        tmp_dq = binned_dq.reshape(binned_ydim_dq // bin_y, bin_y, binned_xdim_dq // bin_x, bin_x)
        tmp_dq2 = np.bitwise_or.reduce(tmp_dq, axis=3)
        binned_dq = np.bitwise_or.reduce(tmp_dq2, axis=1)
        self.dq_image = binned_dq

        # Bin superdark in spatial directions
        for i,sd in enumerate(self.superdarks):
            phastart = self.pha_bins[i]
            phaend = self.pha_bins[i+1]

            binned = sd[b_y0:b_y1, b_x0:b_x1]
            binned_sh = np.shape(binned)
            binned_xdim = binned_sh[1]
            binned_ydim = binned_sh[0]
            tmp = binned.reshape(binned_ydim // bin_y, bin_y, binned_xdim // bin_x, bin_x)
            binned = tmp.sum(axis=3).sum(axis=1)
            self.superdarks[i] = binned
            key = f"pha{phastart}-{phaend}"
            self.pha_images[key] = binned
            rate = binned/self.total_exptime
            zeroinds = np.where(binned == 0)
            nzero = len(zeroinds[0])
            if verbose is True:
                print("")
                print(f"Binning superdark {infile}")
                print(f"Binning PHAs {phastart} through {phaend}")
                print(f"Binning by X={self.bin_x}, Y={self.bin_y}")
                print(f"\tTotal number of events: {np.sum(binned):,}")
                print(f"\tTotal exptime [s] of superdark: {self.total_exptime:,.1f}")
                print(f"\tMinimum number of events in a binned pixel: {np.min(binned)}")
                print(f"\tMaximum number of events in a binned pixel: {np.max(binned)}")
                print(f"\tMean number of events per binned pixel: {np.mean(binned):.1f}")
                print(f"\t  Standard deviation: {np.std(binned):.1f}")
                print(f"\tMedian number of events per binned pixel: {np.median(binned):.1f}")
                print(f"\tMean countrate per binned pixel: {np.mean(rate):.2e}")
                print(f"\t  Standard deviation: {np.std(rate):.2e}")
                print(f"\tMedian countrate per binned pixel: {np.median(rate):.2e}")
                print(f"\tNumber of binned pixels with zero events: {nzero:,}")

        self.is_binned = True 

        if writefile is True:
            self.write_superdark()
        if makeplots is True: 
            self.plot_superdarks()

    def plot_superdarks(self, pdffile=None, vmin=None, vmax=None, savepng=False, pngroot=None):
        if savepng is False:
            if pdffile is None:
                pdffile = os.path.join(self.outdir, self.outfile.replace("asdf", "pdf"))
            pdf = PdfPages(pdffile)
        if savepng is True and pngroot is None:
            pngroot = os.path.join(self.outdir, self.outfile.replace(".asdf", ""))
        for i,sd in enumerate(self.superdarks):
            phastart = self.pha_bins[i]
            phaend = self.pha_bins[i+1]
            rate = sd/self.total_exptime
            fig, ax = plt.subplots(figsize=(20,5))
            #vmin = np.mean(rate) - 3*np.std(rate)
            if vmin is None:
                vmin = np.median(rate) - np.median(rate)*0.5
                if vmin < 0:
                    vmin = 0
            #vmax = np.mean(rate) + 3*np.std(rate)
            if vmax is None:
                vmax = np.median(rate) + np.median(rate)*1.
            im = ax.imshow(rate, aspect="auto", interpolation="nearest", 
                           origin="lower", cmap="inferno", vmin=vmin, vmax=vmax)
            cbar = fig.colorbar(im, label="Counts/s", format="%.1e", pad=0.01)
            cbarticks = cbar.get_ticks()
            # If there are more than 5 tickmarks, reduce it.
            if len(cbarticks) > 7: # 1st and last ticks are not shown
                cbar.ax.locator_params(nbins=5)
            ax.set_title(f"{self.segment}; HV={self.hv}; MJD {self.mjdstarts}-{self.mjdends}; PHA {phastart}-{phaend}; X bin={self.bin_x}, Y bin={self.bin_y}")
            ax.set_xlabel("X (binned)")
            ax.set_ylabel("Y (binned)")
            xticks = ax.xaxis.get_major_ticks()
            xticks[0].set_visible(False)
            xticks[1].set_visible(False)
            if savepng is False:
                pdf.savefig(fig, bbox_inches="tight")
            else:
                pngfile = pngroot+f"_pha{phastart}-{phaend}.png"
                fig.savefig(pngfile, bbox_inches="tight", dpi=200)
                print(f"Wrote {pngfile}")
        if savepng is False:
            pdf.close()
            print(f"Wrote {pdffile}")
        plt.close('all')
        plt.close(fig)


    # Should investigate if this can be replaced with typical sigma clipping
    def screen_counts(self, verbose=True, sigma=10, mask=False, interpolate=False, interp_kernel=None,  method=np.median, exclude_zeros=True):
        for i,sd in enumerate(self.superdarks):
            phastart = self.pha_bins[i]
            phaend = self.pha_bins[i+1]
            key = f"pha{phastart}-{phaend}"
            if exclude_zeros is True:
                nonzeroinds = np.where(sd > 0)
                sd_stats = sd[nonzeroinds]
            else:
                sd_stats = sd
            mid = method(sd_stats)
            std = np.std(sd_stats)
            sigma_cutoff = (std*sigma)
            bad = np.where(sd > (mid+sigma_cutoff))
            if len(bad[0]) == 0:
                continue
            if verbose is True:
                print(f"{len(bad[0])} pixels have counts above {sigma}sigma for {key}")

            if interpolate is True:
                if interp_kernel is None:
                    kernel = Gaussian2DKernel(x_stddev=.3, y_stddev=.3, x_size=11, y_size=5)
                # From https://docs.astropy.org/en/stable/convolution/index.html
                sd[bad] = np.nan
                interp_sd = interpolate_replace_nans(sd, kernel)
                self.superdarks[i] = interp_sd
                self.pha_images[key] = interp_sd
            elif mask is True:
                sd_masked = ma.masked_greater_equal(sd, mid+sigma_cutoff)
                self.superdarks[i] = sd_masked
                self.pha_images[key] = sd_masked
            else:
                sd[bad] = 0.0
                self.superdarks[i] = sd
                self.pha_images[key] = sd


    def write_fits(self, fitsfile=None):
        if fitsfile is None:
            nowdt = datetime.datetime.now()
            now = nowdt.strftime("%d%b%Y")
            fitsfile = self.outfile.replace(".asdf", ".fits")
        self.fitsfile = fitsfile
        if self.overwrite is False and os.path.exists(fitsfile):
            print(f"{fitsfile} already exists and overwrite is False")
            return
        hdr0 = fits.Header()
        hdr0["suprdark"] = os.path.basename(self.outfile)
        hdr0["hv"] = self.hv
        hdr0["segment"] = self.segment 
        primary = fits.PrimaryHDU(header=hdr0)
        hduims = []
        for i,sd in enumerate(self.superdarks):
            hdr = fits.Header()
            hdr["EXTNAME"] = "SCI"
            hdr["phastart"] = self.pha_bins[i]
            hdr["phaend"] = self.pha_bins[i+1]
            hdr["xstart"] = self.xstart
            hdr["ystart"] = self.ystart
            hdr["xend"] = self.xend
            hdr["yend"] = self.yend
            hdr["xbinsize"] = self.bin_x
            hdr["ybinsize"] = self.bin_y
            hdu = fits.ImageHDU(sd, header=hdr)
            hduims.append(hdu)
        hdulist = fits.HDUList([primary] + hduims)
        hdulist.writeto(fitsfile, overwrite=self.overwrite)
        print(f"Wrote {fitsfile}") 


    def write_cube_fits(self, fitsfile=None):
        if fitsfile is None:
            nowdt = datetime.datetime.now()
            now = nowdt.strftime("%d%b%Y")
            fitsfile = self.outfile.replace(".asdf", "_cube.fits")
        self.fitsfile = fitsfile
        if self.overwrite is False and os.path.exists(fitsfile):
            print(f"{fitsfile} already exists and overwrite is False")
            return
        hdr0 = fits.Header()
        hdr0["suprdark"] = os.path.basename(self.outfile)
        hdr0["hv"] = self.hv
        hdr0["segment"] = self.segment 
        primary = fits.PrimaryHDU(header=hdr0)
        cube = np.stack(self.superdarks, axis=0)
        hdr1 = fits.Header()
        hdr1["EXTNAME"] = "SCI"
        hdr1["phastart"] = self.pha_bins[0]
        hdr1["phaend"] = self.pha_bins[-1]
        hdr1["xstart"] = self.xstart
        hdr1["ystart"] = self.ystart
        hdr1["xend"] = self.xend
        hdr1["yend"] = self.yend
        hdr1["xbinsize"] = self.bin_x
        hdr1["ybinsize"] = self.bin_y
        hdu1 = fits.ImageHDU(cube, header=hdr1)
        hdulist = fits.HDUList([primary, hdu1])
        hdulist.writeto(fitsfile, overwrite=self.overwrite)
        print(f"Wrote {fitsfile}") 


    def get_binning_pars(self):
        keys = ["bin_pha", "bin_x", "bin_y", "xstart", "xend", "ystart", "yend",
                "phastart", "phaend"]
        binning = {}
        for k in keys:
            binning[k] = getattr(self, k)

        return binning


def get_gsag_holes(gsagtab, segment, hv):
    extfound = False
    extnum = 1
    hvkey = f"HVLEVEL{segment[-1]}"
    while extfound is False:
        try:
            gsag_hv = fits.getval(gsagtab, hvkey, extnum)
            gsag_seg = fits.getval(gsagtab, "segment", extnum)
        except KeyError:
            extnum += 1
            continue
        if gsag_hv == hv and gsag_seg == segment:
            extfound = True
        else:
            extnum += 1
    gsag = Table.read(gsagtab, format="fits", hdu=extnum)
    gsag_df = gsag.to_pandas()
    gsag_df = gsag_df.rename(columns={"LX": "X0", "LY": "Y0"})
    gsag_df["X1"] = gsag_df["X0"] + gsag_df["DX"]
    gsag_df["Y1"] = gsag_df["Y0"] + gsag_df["DY"]
    gsag_df = gsag_df.astype("int32")

    return gsag_df



import datetime
import os
import glob
import numpy as np
import pandas as pd
from ftplib import FTP
from astropy.io import fits
from astropy.time import Time
from astropy.io import ascii

__author__ = "Jo Taylor"
__email__ = "jotaylor@stsci.edu"

def measure_darkrate(filename, apertures=None):
    """
    For an input dark dataset, record exposure information including
    observation time, observatory latitude & longitude. Measure dark rate at
    specified regions of the COS FUV detector at a given time interval in order
    to accumulate enough counts for a measurement.
    (Taken from the STScI COS team"s dark monitor, dark_monitor.py, and modified. 
    Original authors include Justin Ely, Mees Fix, Dzhuliya Dashtamirova.)
    
    Args:
        filename (str): Name of dark dataset.
        apertures (dict): Output from `get_aperture_region()`. This is passed in to avoid
            running `get_aperture_region()` many times within this function.
   
    Returns:
        dark_df (:obj:`pandas.dataframe`): Pandas dataframe with information for each
            dark measurement. 
    """

    hdulist = fits.open(filename)
    
    if apertures is None:
        apertures = get_aperture_region()

    timeline = hdulist["timeline"].data
    events = hdulist["events"].data
    segment = hdulist[0].header["segment"]
    rootname = hdulist[0].header["rootname"]

    if segment == "FUVA":
        location = {"inner": (1260, 15119, 375, 660), "bottom": (1060, 15250, 296, 375),
                    "top": (1060, 15250, 660, 734), "left": (1060, 1260, 296, 734),
                    "right": (15119, 15250, 296, 734)}
        location.update(apertures[segment])
    elif segment == "FUVB":
        location = {"inner": (1000, 14990, 405, 740), "bottom": (809, 15182, 360, 405),
                    "top": (809, 15182, 740, 785), "left": (809, 1000, 360, 785),
                    "right": (14990, 15182, 360, 785)}
        location.update(apertures[segment])
    pha = (2, 23)
    timestep = 25 # in units of seconds
    times = timeline["time"][::timestep].copy()
    events = hdulist["events"].data

    lat = timeline["latitude"][::timestep][:-1].copy()
    lon = timeline["longitude"][::timestep][:-1].copy()
    second_per_mjd = 1.15741e-5
    mjd_per_step = hdulist[1].header["EXPSTART"] + times.copy().astype(np.float64) * second_per_mjd
    mjd = mjd_per_step[:-1]                                                   
    t = Time(mjd, format="mjd") 
    mjdtime = [x.value for x in t]
    
    # Creates different tables, one for each region of detector
    info_dict = {"segment": segment, "rootname": rootname,"latitude": lat, 
                 "longitude": lon, "time": mjdtime, "pha": np.arange(32), 
                 "exptime": timestep}
    dark_dict = {}
    for x in location:
        region_area = (location.get(x)[1] - location.get(x)[0]) * (location.get(x)[3] - location.get(x)[2])
        # In reality this is not an issue since YCORR=YFULL for darks.
        if "location" == "apertures":
            coord_x = "XCORR"
            coord_y = "YFULL"
        else:
            coord_x = "XCORR"
            coord_y = "YCORR"
        counts_pha = []
        for phabin in range(32):
            index = np.where((events["PHA"] == phabin) &
                             (events[coord_x] > location.get(x)[0]) &
                             (events[coord_x] < location.get(x)[1]) &
                             (events[coord_y] > location.get(x)[2]) &
                             (events[coord_y] < location.get(x)[3]))
            counts = np.histogram(events[index]["time"], bins=times)[0]
            counts_pha.append(counts)
        counts_time = np.dstack(counts_pha)[0]
        region_d = {"darks": counts_time, "xcorr_min": location[x][0], 
                    "xcorr_max": location[x][1], "ycorr_min": location[x][2], 
                    "ycorr_max": location[x][3], "region_area": region_area}
        dark_dict[x] = region_d
    hdulist.close() 
    
    return info_dict, dark_dict

def get_solar_data(solardir):
    """
    Pull solar data files from NOAA website. Solar data is FTPd from NOAA and 
    written to text files.
    (Taken from the STScI COS team"s dark monitor, dark_monitor.py, and modified. 
    Original authors include Justin Ely, Mees Fix, Dzhuliya Dashtamirova.)

    Args:
        solardir (str): Directory to write the solar data files to.
    """

    ftp = FTP("ftp.swpc.noaa.gov")
    ftp.login()
    ftp.cwd("/pub/indices/old_indices/")

    for item in sorted(ftp.nlst()):
        if item.endswith("_DSD.txt"):
            year = int(item[:4])
            if year >= 2008:
                print("Retrieving: {}".format(item))
                destination = os.path.join(solardir, item)
                ftp.retrbinary("RETR {}".format(item),
                               open(destination, "wb").write)
                os.chmod(destination, 0o777)


def to_mjd(row):
    fulltime = row["time-tag"]+"-01 00:00:00"
    mjd = Time(fulltime, scale="utc", format="iso").mjd
    return mjd


def parse_solar_json(infile):
    """
    Parse new NOAA solar data. Taken from:
    https://www.swpc.noaa.gov/products/solar-cycle-progression
    which leads to:
    https://services.swpc.noaa.gov/json/solar-cycle/observed-solar-cycle-indices.json
    """
    df = pd.read_json(infile)
    df = df.drop(["ssn", "smoothed_ssn", "observed_swpc_ssn", "smoothed_swpc_ssn"], axis=1)
    df = df.loc[df["time-tag"] >= '2009-01']
    df["mjd"] = df.apply(to_mjd, axis=1) 
    df = df.drop(["time-tag"], axis=1)
    return df["mjd"].values, df["f10.7"].values


def parse_solar_files(files):
    """
    Parse solar data text files and return date and flux.
    (Taken from the STScI COS team"s dark monitor, dark_monitor.py, and modified. 
    Original authors include Justin Ely, Mees Fix, Dzhuliya Dashtamirova.)
    
    Args:
        files (str or array-like): If a string, a check is done to see if string 
            is a file or a directory to be globbed. If not a string, array-like
            is assumed.
    
    Returns:
        date (:obj:`numpy.ndarray`): MJD of each solar flux measurement.
        flux (:obj:`numpy.ndarray`): Solar flux measurements.
    """

    date = []
    flux = []
    if isinstance(files, str):
        if os.path.isfile(files):
            input_list = [files]
        else:
            input_list = glob.glob(os.path.join(files, "*DSD.txt"))
    else:
        input_list = files
    input_list.sort()

    for item in input_list:
#        print("Reading {}".format(item))

        # clean up Q4 files when year-long file exists
        if ("Q4_" in item) and os.path.exists(item.replace("Q4_", "_")):
            print("Removing duplicate observations: {}".format(item))
            os.remove(item)
            continue

        # astropy.ascii no longer returns an empty table for empty files
        # Throws IndexError, we will go around it if empty.
        try:
            data = ascii.read(item, data_start=1, comment="[#,:]")
        except IndexError:                                    
            continue                                          
                                                              
        for line in data:                                     
            line_date = Time("{}-{}-{} 00:00:00".format(line["col1"],       
                                                        line["col2"],       
                                                        line["col3"]),      
                             scale="utc", format="iso").mjd   
            line_flux = line[3]                               
                                                              
            if line_flux > 0:                                 
                date.append(line_date)                        
                flux.append(line_flux)                        
                                                              
    return np.array(date), np.array(flux)                     

def get_aperture_region(cenwave=1291, aperture="PSA", segments=["FUVA", "FUVB"], life_adj=[1, 2, 3, 4, 5, 6]):
    """
    Determine the extraction box for a given cenwave (1291 by default) and
    aperture (PSA by default).
    
    Args:
        cenwave (int): Cenwave to look up.
        aperture (str): Aperture to look up.

    Returns:
        cenwave (int): Cenwave setting.
        apertures (dict): Dictionary where each key is segment, and the value
            is another dictionary where each key is lifetime position,
            and the value is a tuple with (xmin, xmax, ymin, ymax) representing
            the cenwave extraction box for that segment and LP combo.
    """    

    if isinstance(life_adj, int):
        life_adj = [life_adj] 
    today0 = datetime.datetime.now()
    today = today0.strftime("%Y-%m-%d")

    if "CRDS_PATH" not in os.environ:
        if os.path.exists("/grp/crds/cache"):
            os.environ["CRDS_PATH"] = "/grp/crds/cache"
        else:
            raise AssertionError("CRDS_PATH environment variable must first be defined")
    os.environ["CRDS_SERVER_URL"] = "https://hst-crds.stsci.edu"
    import crds
    current_pmap = crds.get_default_context() 

    # The active area limits are taken from the COS BRFTAB x1u1459il_brf.fits
    # This file will almost certainly not be updated, so hardcoding is okay.
    aa_xcorr = {"FUVA": [1060, 15250], "FUVB": [809, 15182]}
    # aa_ycorr = {"FUVA": [296, 734], "FUVB": [360, 785]}

    apertures = {"FUVA": {}, "FUVB": {}}
    # For each LP, determine the appropriate xtractab as returned by CRDS on the fly.
    for segment in segments:
#        for life_adj, date_obs in zip([1, 2, 3, 4], ["2010-01-01", "2014-01-01", "2016-01-01", "2018-01-01"]):
        for lp in life_adj:
            crds_1dx = crds.getrecommendations(parameters={"INSTRUME": "COS", 
                                "DETECTOR": "FUV", "LIFE_ADJ": lp, 
                                "OBSTYPE": "SPECTROSCOPIC", 
                                "DATE-OBS": today, "TIME-OBS": "00:00:00",
                                "CENWAVE": cenwave},
                            reftypes=["xtractab"], context=current_pmap, observatory="hst")
            lp_1dx = os.path.join(os.environ["CRDS_PATH"], "references/hst/", 
                                  crds_1dx["xtractab"])
    
            data_1dx = fits.getdata(lp_1dx, 1)
            ind = np.where((data_1dx["cenwave"] == cenwave) & 
                           (data_1dx["segment"] == segment) & 
                           (data_1dx["aperture"] == aperture))
            aperture_data = data_1dx[ind]
    
            x = np.arange(16384)
            y_center = aperture_data["slope"][0] * x + aperture_data["b_spec"][0]
            y_upper = round(max(y_center + aperture_data["height"][0] / 2))
            y_lower = round(min(y_center - aperture_data["height"][0] / 2))
    
            lpkey = f"lp{lp}_{aperture.lower()}_{cenwave}"

            # This  matches the format already defined in measure_darkrate()
            # i.e. xmin, xmax, ymin, ymax 
            # VERY IMPORTANT: Y coords are in YFULL, X in XCORR
            apertures[segment][lpkey] = (aa_xcorr[segment][0], aa_xcorr[segment][1],
                                        int(y_lower), int(y_upper))
    return apertures
